/*
	ideturbo.c - IDE drive streaming I/O support, as used in the
	TiVo Personal Video Recorder system.
	
	Copyright TiVo Inc.,  1998, 1999, 2000, 2001.  All Rights Reserved
	Worldwide.
	
	This source code is being made available under the terms of the
	GNU General Public License, version 2 (July 1991).  The complete
	text of this license may be found on the Free Software Foundation
	web site at http://www.fsf.org/
*/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/timer.h>
#include <linux/mm.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>
#include <linux/major.h>
#include <linux/blkdev.h>
#include <linux/errno.h>
#include <linux/hdreg.h>
#include <linux/genhd.h>
#include <linux/malloc.h>

#include <asm/byteorder.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/io.h>

#include <linux/ide.h>
#include <linux/ide-tivo.h>
#include <linux/tivo-private.h>

// #define DUMP_CONFIG
// #define WATCH_FLUSHES

static int swaps = 0;
static int flushes = 0;

int cautionlevel = 3;

extern int (*idetivo_statistics_hook)(ide_drive_t *drive,
				      struct idetivo_request *req,
				      int reset_statistics,
				      struct request *rq);

extern void (*idetivo_cmd_hook)(ide_drive_t *drive, unsigned char *cmdcode)
     __attribute__ ((weak));

extern void idetivo_insert_request(ide_drive_t *drive, struct idetivo_request *req, int athead);

static void idetivo_bswap_data (void *buffer, int wcount)
{
    u16 *p = buffer;
   
    //    printk(KERN_INFO "Byteswapping %d words\n", wcount);

    while (wcount--) {
        *p++ = *p << 8 | *p >> 8;
        *p++ = *p << 8 | *p >> 8;
    }
}

static inline void idedisk_input_data (ide_drive_t *drive, void *buffer, unsigned int wcount)
{
	ide_input_data(drive, buffer, wcount);
#ifdef CONFIG_TCD
	idetivo_bswap_data(buffer, wcount);
#endif
}

static inline void idedisk_output_data (ide_drive_t *drive, void *buffer, unsigned int wcount)
{
#ifdef CONFIG_TCD
  idetivo_bswap_data(buffer, wcount);
#endif
  ide_output_data(drive, buffer, wcount);
#ifdef CONFIG_TCD
  idetivo_bswap_data(buffer, wcount);
#endif
}

static void reissue(ide_drive_t *drive) {
   struct idetivo_request *req;
   long int flags;
   if (IDE_TIVO_INFO && (req = IDE_TIVO_INFO->cur)) {
#ifdef DUMP_CONFIG
      printk(KERN_INFO "Reissue TiVo request on %s\n", drive->name);
#endif
      // put it back on the head of both request queues!
      IDE_TIVO_INFO->cur = NULL;
      idetivo_insert_request(drive, req, 1);
   } else {
#ifdef DUMP_CONFIG
      printk(KERN_INFO "Retrigger standard request on %s\n", drive->name);
#endif
      blk_dev[HWIF(drive)->major].current_request = HWGROUP(drive)->rq = NULL;
   }
}

/*
 * getconfig_intr() is invoked on completion of a getconfig cmd.
 */
static void getconfig_intr (ide_drive_t *drive)
{
   struct idetivo_drive_info *tivodisk = (struct idetivo_drive_info *) drive->driver_data;
   byte stat = GET_STAT();

#ifdef DUMP_CONFIG
   printk(KERN_INFO "In getconfig_intr\n");
#endif
   if (!OK_STAT(stat,DATA_READY,BAD_STAT)) {
      ide_error(drive, "getconfig_intr", stat);
   } else {
#ifdef DUMP_CONFIG
     printk("Getting config data from drive\n");
#endif
      idedisk_input_data(drive, tivodisk->config, SECTOR_WORDS);
#ifdef DUMP_CONFIG
      printk("Read %d words to 0x%p\n", SECTOR_WORDS, tivodisk->config);
#endif
      tivodisk->state = STATE_KNOWN;
   }
   reissue(drive);
}

/*
 * setconfig_intr() is invoked on completion of a setconfig cmd.
 */
static void setconfig_intr (ide_drive_t *drive)
{
    byte stat = GET_STAT();

#ifdef DUMP_CONFIG
    printk(KERN_INFO "In setconfig_intr\n");
#endif
    if (!OK_STAT(stat,READY_STAT,BAD_STAT)) {
       ide_error(drive, "setconfig_intr", stat);
    } else {
#ifdef DUMP_CONFIG
      printk(KERN_INFO "Sent config info to %s\n", drive->name);
#endif
    }
   reissue(drive);
}

/*
 * flush_intr() is invoked on completion of a flush cmd.
 */
static void flush_intr (ide_drive_t *drive)
{
    byte stat = GET_STAT();

#ifdef DUMP_CONFIG
    printk(KERN_INFO "In flush_intr\n");
#endif
    if (!OK_STAT(stat,READY_STAT,BAD_STAT)) {
      struct idetivo_drive_info *tivodisk = (struct idetivo_drive_info *)
	drive->driver_data;
      ide_error(drive, "flush_intr", stat);
      if (tivodisk && tivodisk->family == FAMILY_UNKNOWN) {
	printk(KERN_ERR "Drive %s doesn't implement FLUSH command??\n",
	       drive->name);
	tivodisk->state = STATE_NOMEDIA;
      }
    } else {
#ifdef DUMP_CONFIG
      printk(KERN_INFO "Flushed buffer on %s\n", drive->name);
#endif
    }
   reissue(drive);
}

/*
 * misc_intr() is invoked on completion of miscellaneous cmds.
 */
static void misc_intr (ide_drive_t *drive)
{
    byte stat = GET_STAT();

#ifdef DUMP_CONFIG
    printk(KERN_INFO "In misc_intr\n");
#endif
    if (!OK_STAT(stat,READY_STAT,BAD_STAT)) {
       ide_error(drive, "misc_intr", stat);
    }
    reissue(drive);
}

static void initiate_flush (ide_drive_t *drive)
{
   unsigned long flags;
   save_flags(flags);
   sti();
     {
	byte stat;
	stat = GET_STAT();
	if (stat & BUSY_STAT) {
	   printk(KERN_INFO "ideturbo: initiate_flush with drive in use, stat 0x%X\n", stat);
	}
     }
   OUT_BYTE(drive->ctl,IDE_CONTROL_REG);
   OUT_BYTE(drive->select.b.unit ? 0xB0 : 0xA0, IDE_SELECT_REG); /* drive */
   ide_set_handler (drive, &flush_intr, WAIT_CMD);
#ifdef DUMP_CONFIG
   printk("ideturbo: flushing buffer on %s\n", drive->name);
#endif
   OUT_BYTE(0xE7, IDE_COMMAND_REG);
   restore_flags(flags);
#ifdef WATCH_FLUSHES
   flushes++;
   if ((flushes % 100) == 0) {
      printk("ideturbo: %d flushes\n", flushes);
   }
#endif   
}

static int ide_genericstats(ide_drive_t *drive,
			    struct idetivo_request *req,
			    int reset_statistics,
			    struct request *rq)
{
   struct idetivo_drive_info *tivodisk =
     (struct idetivo_drive_info *) drive->driver_data;
   long int flags;
   int theCmd;
   if (reset_statistics) {
      tivodisk->state = STATE_UNKNOWN;
      return 0;
   }
   switch (tivodisk->state) {
   case STATE_UNKNOWN:
     tivodisk->back_to_back_limit = 1;
     tivodisk->state = STATE_MEDIA;
     /* drop through */
   case STATE_MEDIA:
     if (rq) {
       theCmd = rq->cmd;
     } else {
       theCmd = req->cmd;
     }
     if (theCmd == READ) {
       tivodisk->back_to_back_writes = 0;
     } else {
       tivodisk->back_to_back_writes ++;
       if (tivodisk->back_to_back_writes > tivodisk->back_to_back_limit) {
	 tivodisk->back_to_back_writes = 0;
	 initiate_flush(drive);
	 return 1;
       }
     }
     return 0;
   case STATE_NOMEDIA:
   default:
     return 0;
   }
}

static int ide_tivoavstats(ide_drive_t *drive,
			   struct idetivo_request *req,
			   int reset_statistics,
			   struct request *rq)
{
   struct idetivo_drive_info *tivodisk =
     (struct idetivo_drive_info *) drive->driver_data;
   long int flags;
   int theCmd;
   if (reset_statistics) {
      tivodisk->state = STATE_UNKNOWN;
      return 0;
   }
   switch (tivodisk->state) {
   case STATE_UNKNOWN:
     tivodisk->back_to_back_limit = 1;
     tivodisk->state = STATE_NOMEDIA;
     printk(KERN_INFO "%s: Enabling write cache\n", drive->name);
     save_flags(flags);
     sti();
     {
       byte stat;
       stat = GET_STAT();
       if (stat & BUSY_STAT) {
	 printk(KERN_INFO "ideturbo: entered with drive in use, stat 0x%X\n",
		stat);
       }
     }
     OUT_BYTE(drive->ctl,IDE_CONTROL_REG);
     OUT_BYTE(drive->select.b.unit ? 0x10 : 0x00, IDE_SELECT_REG); /* drive */
     ide_set_handler (drive, &misc_intr, WAIT_CMD);
     OUT_BYTE(0x02, IDE_FEATURE_REG);
     OUT_BYTE(0xEF, IDE_COMMAND_REG);
     restore_flags(flags);
     return 1; /* wait for interrupt before proceeding */
    case STATE_NOMEDIA:
     memset(tivodisk->config, 0, sizeof tivodisk->config);
     /*
       Change the following as necessary.  Note that the WD drives
       organize data in terms of shortwords, not bytes (I missed this
       initially!).  Assume big-endian byte ordering.
     */
     tivodisk->config[1] =  1;
     tivodisk->config[9] =  255;
     tivodisk->config[11] =  255;
     tivodisk->config[13] =  2;
     /*
       See above
     */
     printk(KERN_INFO "%s: Setting new A/V configuration:\n", drive->name);
     printk(KERN_INFO " A/V features enabled:      %d\n",
	    tivodisk->config[1]&1);
     printk(KERN_INFO " A/V read recovery time:    %d msec\n",
	    tivodisk->config[9]);
     printk(KERN_INFO " A/V write recovery time:   %d msec\n",
	    tivodisk->config[11]);
     printk(KERN_INFO " A/V max streams:           %d %s\n",
	    tivodisk->config[13] & 0x7,
	    (tivodisk->config[13] & 0x7) ? "" : "unlimited");
     tivodisk->state = STATE_MEDIA;
     save_flags(flags);
     sti();
     {
       byte stat;
       stat = GET_STAT();
       if (stat & BUSY_STAT) {
	 printk(KERN_INFO "ideturbo: entered with drive in use, stat 0x%X\n",
		stat);
       }
     }
     OUT_BYTE(drive->ctl,IDE_CONTROL_REG);
     OUT_BYTE(drive->select.b.unit ? 0x10 : 0x00, IDE_SELECT_REG); /* drive */
     ide_set_handler (drive, &setconfig_intr, WAIT_CMD);
     OUT_BYTE(0x80, IDE_COMMAND_REG);
     if (ide_wait_stat(drive, DATA_READY, drive->bad_wstat, WAIT_DRQ)) {
       printk(KERN_ERR "%s: no DRQ after issuing set-configuration command\n",
	      drive->name);
       return 0;
     }
     idedisk_output_data(drive, tivodisk->config, SECTOR_WORDS);
     restore_flags(flags);
     return 1; /* wait for interrupt before proceeding */
   case STATE_MEDIA:
   default:
     if (rq) {
       theCmd = rq->cmd;
     } else {
       theCmd = req->cmd;
     }
     if (theCmd == READ) {
       tivodisk->back_to_back_writes = 0;
     } else {
       tivodisk->back_to_back_writes ++;
       if (tivodisk->back_to_back_writes > tivodisk->back_to_back_limit) {
	 tivodisk->back_to_back_writes = 0;
	 initiate_flush(drive);
	 return 1;
       }
     }
     return 0;
   }
}

int ide_tivostats(ide_drive_t *drive,
		  struct idetivo_request *req,
		  int reset_statistics,
		  struct request *rq)
{
   struct idetivo_drive_info *tivodisk = (struct idetivo_drive_info *)
     drive->driver_data;
   enum media_state desired_state;
   int theCmd;
#ifdef DUMP_CONFIG
   char strbuf[80];
   int i, j;
   char hexen[] = "0123456789ABCDEF";
#endif
   long int flags;
//   printk(KERN_INFO "ide_tivostats entered for %s, req is %p\n", drive->name, req);
   if (IDE_TIVO_INFO->version != IDETIVO_DRIVER_VERSION) {
     printk(KERN_ERR "ideturbo version mismatch, disabling\n");
     idetivo_statistics_hook = NULL;
     idetivo_cmd_hook = NULL;
     return 0;
   }
   if (!IDE_TIVO_DRIVE_INFO) {
     return 0;
   }
   if (IDE_TIVO_DRIVE_INFO->family == FAMILY_UNKNOWN) {
     if (drive->id->model[0] == 'Q' &&
	 drive->id->model[1] == 'U' &&
	 drive->id->model[2] == 'A') {
       printk(KERN_INFO "%s: Quantum management\n", drive->name);
       IDE_TIVO_DRIVE_INFO->family = FAMILY_QUANTUM;
     } else if (drive->id->model[0] == 'W' &&
		drive->id->model[1] == 'D' &&
		drive->av_cmdset) {
       printk(KERN_INFO "%s: StreamWeaver management\n", drive->name);
       IDE_TIVO_DRIVE_INFO->family = FAMILY_STREAMWEAVER1;
     } else {
       printk(KERN_INFO "%s: Generic ATA management\n", drive->name);
       IDE_TIVO_DRIVE_INFO->family = FAMILY_UNSUPPORTED;
     }
   }
   
   switch (IDE_TIVO_DRIVE_INFO->family) {
   case FAMILY_UNSUPPORTED:
     return ide_genericstats(drive, req, reset_statistics, rq);
   case FAMILY_STREAMWEAVER1:
     return ide_tivoavstats(drive, req, reset_statistics, rq);
   case FAMILY_QUANTUM:
   default:
     break;
   }
     
   if (reset_statistics) {
      tivodisk->state = STATE_UNKNOWN;
      return 0;
   }
   if (tivodisk->state == STATE_UNKNOWN) {
      tivodisk->back_to_back_limit = 1;
      printk(KERN_INFO "Need to get drive statistics\n");
      save_flags(flags);
      sti();
      OUT_BYTE(drive->ctl,IDE_CONTROL_REG); /* turn off interrupt */
      OUT_BYTE(drive->select.b.unit ? 0xB0 : 0xA0, IDE_SELECT_REG); /* drive */
      OUT_BYTE(0x01, IDE_NSECTOR_REG); /* get config, no save */
      OUT_BYTE(0xFF, IDE_SECTOR_REG);
      OUT_BYTE(0xFF, IDE_LCYL_REG);
      OUT_BYTE(0x3F, IDE_HCYL_REG);
      ide_set_handler (drive, &getconfig_intr, WAIT_CMD);
      OUT_BYTE(0xF0, IDE_COMMAND_REG);
      restore_flags(flags);
      return 1; /* go around again... */
   }
#if 0
   if (req && (req->partition_flags & 1)) {
      desired_state = STATE_MEDIA;
   } else {
      desired_state = STATE_NOMEDIA;
   }
#else
   /* Force application mode. */
   desired_state = STATE_NOMEDIA;
#endif
   if (rq) {
     theCmd = rq->cmd;
   } else {
     theCmd = req->cmd;
   }
   if (theCmd == READ) {
     tivodisk->back_to_back_writes = 0;
   } else {
     tivodisk->back_to_back_writes ++;
     if (tivodisk->back_to_back_writes > tivodisk->back_to_back_limit) {
       tivodisk->back_to_back_writes = 0;
       initiate_flush(drive);
       return 1;
     }
   }
   if (tivodisk->state == desired_state) {
      return 0;
   }
   if (tivodisk->state != STATE_FLUSHED) {
      initiate_flush(drive);
      tivodisk->state = STATE_FLUSHED;
      return 1;
   }
#ifdef DUMP_CONFIG
   printk (KERN_INFO "Switching %s to %s mode\n", drive->name,
	   (desired_state == STATE_MEDIA) ? "media" : "application");
#endif
   tivodisk->config[32] = 0x3; /* prefetch enable, read cache enable */
   switch (desired_state) {
    case STATE_MEDIA:
      switch (cautionlevel) {
      case 0:
	tivodisk->config[36] = (tivodisk->config[36] & 0x26) | 0x10;
	tivodisk->config[37] = 0;
	break;
      case 1:
	tivodisk->config[36] = (tivodisk->config[36] & 0x26) | 0x08;
	tivodisk->config[37] = 0;
	break;
      case 2:
	tivodisk->config[36] = (tivodisk->config[36] & 0x26) | 0x08;
	tivodisk->config[37] = 1;
	break;
      default:
	tivodisk->config[36] = (tivodisk->config[36] & 0x26) | 0x08;
	tivodisk->config[37] = 2;
	break;
      }
      tivodisk->config[39] = (tivodisk->config[39] & 0xFD) | 0x4;
      break;
    case STATE_NOMEDIA:
      tivodisk->config[36] = (tivodisk->config[36] & 0xEE) | 0xC8;
      tivodisk->config[37] = 2;
      tivodisk->config[39] = tivodisk->config[39] | 0x6;
      break;
    default:
      break;
   }
#ifdef DUMP_CONFIG
   for (i = 0, j = 0 ; i < 64 ; i++) {
      if ((i % 16) == 0) {
	 if (i > 0) {
	    strbuf[j]=0;
	    printk (KERN_INFO "%s\n", strbuf);
	 }
	 j=0;
      }
      strbuf[j++] = hexen[0xF & (tivodisk->config[i] >> 4)];
      strbuf[j++] = hexen[0xF & tivodisk->config[i]];
      strbuf[j++] = ' ';
   }
   strbuf[j]=0;
   printk (KERN_INFO "%s\n", strbuf);
#endif
#ifdef PERIODIC_MODE_LOG
   if ((++swaps % 1000) == 0) {
      printk(KERN_INFO "ideturbo: %d mode changes\n", swaps);
   }
#endif
   save_flags(flags);
   sti();
     {
	byte stat;
	stat = GET_STAT();
	if (stat & BUSY_STAT) {
	   printk(KERN_INFO "ideturbo: entered with drive in use, stat 0x%X\n", stat);
	}
     }
   OUT_BYTE(drive->ctl,IDE_CONTROL_REG);
   OUT_BYTE(drive->select.b.unit ? 0xB0 : 0xA0, IDE_SELECT_REG); /* drive */
   OUT_BYTE(0xFE, IDE_NSECTOR_REG); /* set config, no save */
   OUT_BYTE(0xFF, IDE_SECTOR_REG);
   OUT_BYTE(0xFF, IDE_LCYL_REG);
   OUT_BYTE(0x3F, IDE_HCYL_REG);
   ide_set_handler (drive, &setconfig_intr, WAIT_CMD);
   OUT_BYTE(0xF0, IDE_COMMAND_REG);
   if (ide_wait_stat(drive, DATA_READY, drive->bad_wstat, WAIT_DRQ)) {
      printk(KERN_ERR "%s: no DRQ after issuing set-configuration command\n", drive->name);
      return 0;
   }
   #ifdef NOTDEF
   if (!drive->unmask)
     cli();
   #endif
   idedisk_output_data(drive, tivodisk->config, SECTOR_WORDS);
   tivodisk->state = desired_state;
#ifdef DUMP_CONFIG
   printk (KERN_INFO "Sent data to switch to %s mode\n", drive->name,
	   (desired_state == STATE_MEDIA) ? "media" : "application");
#endif
   restore_flags(flags);
   return 1; /* wait for interrupt before proceeding */
}

void ide_tivocmdhook(ide_drive_t *drive, unsigned char *cmdcode)
{
#ifdef CONFIG_TCD   
  struct idetivo_request *req; 
  static int mutate = 0, total = 0;
  if (IDE_TIVO_INFO->version != IDETIVO_DRIVER_VERSION) {
    printk(KERN_ERR "ideturbo version mismatch, disabling\n");
    idetivo_statistics_hook = NULL;
    idetivo_cmd_hook = NULL;
    return;
  }
  if (IDE_TIVO_DRIVE_INFO->family != FAMILY_STREAMWEAVER1) {
    return;  /* No A/V command set, don't frob command code */
  }
  total++;
#ifdef NOTDEF
  if ((total % 1000) == 0) {
    printk (KERN_INFO "A/V mode: %d total, %d mutations\n", total, mutate);
  }
#endif
  req = IDE_TIVO_INFO->cur;
  if (!req) {
    return;    /* Not a readsectors/writesectors, don't frob command code */
  }
  if (!(req->partition_flags & 1)) {
    return;  /* Not on a media partition, don't frob command code */
  }
  /* 
     TiVo I/O request, on a media partition, on a drive which supports
     the A/V stream commands.  Transform normal read/write command code
     into equivalent media command.  Set the IDE feature register to
     indicate a non-urgent, non-read-continuous request.  Use a stream
     ID of 0 for reads, and a stream ID of 1 for writes.
  */
  mutate ++;
  switch (*cmdcode) {
  case WIN_READ:
    *cmdcode = WIN_READAV;
    OUT_BYTE(0x00, IDE_FEATURE_REG);
    break;
  case WIN_READDMA:
    *cmdcode = WIN_READAVDMA;
    OUT_BYTE(0x00, IDE_FEATURE_REG);
    break;
  case WIN_WRITE:
    *cmdcode = WIN_WRITEAV;
    OUT_BYTE(0x01, IDE_FEATURE_REG);
    break;
  case WIN_WRITEDMA:
    *cmdcode = WIN_WRITEAVDMA;
    OUT_BYTE(0x01, IDE_FEATURE_REG);
    break;
  default:
    break;
  }
#else
  (void) drive;
  (void) cmdcode;
#endif     
}


int init_module (void)
{
   MOD_INC_USE_COUNT;
   idetivo_statistics_hook = ide_tivostats;
   if (&idetivo_cmd_hook != NULL) {
     idetivo_cmd_hook = ide_tivocmdhook;
   }
#ifdef CONFIG_TCD
   printk(KERN_INFO "TiVo disk statistics module loaded (byteswap enabled)\n");
#else
   printk(KERN_INFO "TiVo disk statistics module loaded\n");
#endif
   printk(KERN_INFO "Disk cautionlevel is %d\n", cautionlevel);
   return 0;
}

void cleanup_module (void)
{
   idetivo_statistics_hook = NULL;
   if (&idetivo_cmd_hook != NULL) {
     idetivo_cmd_hook = NULL;
   }
   printk(KERN_INFO "TiVo disk statistics module unloaded\n");
   MOD_DEC_USE_COUNT;
}
